import os
import logging
from glob import glob
from typing import Optional, List, Dict

import gevent
import shutil
from ewoksjob.client import submit

try:
    from bliss import current_session
except ImportError:
    current_session = None

from ..utils import trigger
from ..utils import directories
from ..flint import WithFlintAccess
from ..resources import resource_filename
from ..persistent.parameters import ParameterInfo
from ..persistent.parameters import WithPersistentParameters
from .plots import TemperaturePlot

logger = logging.getLogger(__name__)


class Id24TemperaturePlotter(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("_enabled"),
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("workflow_with_fit", category="workflows"),
        ParameterInfo("energy_name", category="XAS"),
        ParameterInfo("mu_name", category="XAS"),
        ParameterInfo("extend_plotrange_left", category="plot", doc="(nm)"),
        ParameterInfo("extend_plotrange_right", category="plot", doc="(nm)"),
        ParameterInfo("two_color_difference", category="plot", doc="(nm)"),
        ParameterInfo("dpi", category="plot"),
        ParameterInfo("refit", category="fit"),
        ParameterInfo("wavelength_min", category="fit", doc="(nm)"),
        ParameterInfo("wavelength_max", category="fit", doc="(nm)"),
    ],
):
    DEFAULT_WORKFLOW = resource_filename("id24", "id24_planck_plot.json")
    DEFAULT_WORKFLOW_WITH_FIT = resource_filename("id24", "id24_planck_fitplot.json")

    def __init__(self, **defaults) -> None:
        defaults.setdefault("_enabled", False)
        defaults.setdefault("workflow", self.DEFAULT_WORKFLOW)
        defaults.setdefault("workflow_with_fit", self.DEFAULT_WORKFLOW_WITH_FIT)

        defaults.setdefault("energy_name", "energy_enc")
        defaults.setdefault("mu_name", "mu_trans")

        defaults.setdefault("extend_plotrange_left", -15)
        defaults.setdefault("extend_plotrange_right", 50)
        defaults.setdefault("two_color_difference", 42)
        defaults.setdefault("dpi", 150)

        defaults.setdefault("refit", False)
        defaults.setdefault("wavelength_min", None)
        defaults.setdefault("wavelength_max", None)

        super().__init__(**defaults)
        self.plotter = Plotter()
        self._sync_scan_metadata()

    def enable(self) -> None:
        self._enabled = True
        self._sync_scan_metadata()

    def disable(self) -> None:
        self._enabled = False
        self._sync_scan_metadata()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        ntasks = self.plotter.purge_tasks()
        categories["status"] = {
            "Enabled": self._enabled,
            "Fitting": self.refit,
            "Plotting tasks": ntasks,
        }
        return categories

    def _sync_scan_metadata(self) -> None:
        if self._enabled:
            workflows_category = trigger.register_workflow_category(timing="END")
            workflows_category.set("temperature", self.trigger_workflow_for_scan)
        else:
            trigger.unregister_workflow_category()

    def trigger_workflow_for_scan(self, scan) -> Optional[dict]:
        if not self._scan_requires_workflow(scan):
            return
        filename = scan.scan_info["filename"]
        scan_number = scan.scan_info["scan_nb"]
        self.trigger_workflow(scan_number, filename=filename)

    def trigger_workflow(
        self, scan_number: int, filename: Optional[str] = None
    ) -> None:
        if filename:
            if not os.path.isabs(filename):
                basename = os.path.basename(filename)
                raw_data = directories.get_raw_dir(current_session.scan_saving.filename)
                filenames = glob(os.path.join(raw_data, "*", "*", basename))
                if not filenames:
                    raise FileNotFoundError(filename)
                filename = filenames[0]
        else:
            filename = current_session.scan_saving.filename
        workflow = self._get_workflow(filename)

        future = submit(
            args=(workflow,),
            kwargs={
                "inputs": self._get_workflow_inputs(filename, scan_number),
                "outputs": [{"all": False}],
            },
        )
        self.plotter.handle_workflow_result(future)

    def _scan_requires_workflow(self, scan) -> bool:
        if not self._has_temperature(scan):
            return False
        if not scan.scan_info.get("save"):
            return False
        if not scan.scan_info.get("filename"):
            return False
        if not scan.scan_info.get("scan_nb"):
            return False
        return True

    def _has_temperature(self, scan) -> bool:
        channels = scan.scan_info.get("channels", dict())
        return "laser_heating_down:T_planck" in channels

    def _get_workflow(self, filename) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        if self.refit:
            src_file = self.workflow_with_fit
        else:
            src_file = self.workflow
        if src_file is None:
            return
        if not os.path.isfile(src_file):
            if self.refit:
                src_file = self.DEFAULT_WORKFLOW_WITH_FIT
            else:
                src_file = self.DEFAULT_WORKFLOW

        workflow_directory = directories.get_workflows_dir(filename)
        dst_file = os.path.join(workflow_directory, os.path.basename(src_file))
        if src_file != dst_file:
            if self.refit:
                self.workflow_with_fit = dst_file
            else:
                self.workflow = dst_file

        if not os.path.exists(dst_file):
            os.makedirs(workflow_directory, exist_ok=True)
            shutil.copyfile(src_file, dst_file)

        return dst_file

    def _get_workflow_inputs(self, filename: str, scan_number: int) -> List[dict]:
        inputs = self._get_read_inputs(filename, scan_number)
        inputs += self._get_plot_inputs(filename)
        inputs += self._get_fit_inputs()
        return inputs

    def _get_read_inputs(self, filename: str, scan_number: int) -> List[dict]:
        task_identifier = "XasTemperatureRead"
        return [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "scan_number",
                "value": scan_number,
            },
            {
                "task_identifier": task_identifier,
                "name": "retry_timeout",
                "value": 60,
            },
            {
                "task_identifier": task_identifier,
                "name": "energy_name",
                "value": self.energy_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "mu_name",
                "value": self.mu_name,
            },
        ]

    def _get_plot_inputs(self, filename: str) -> List[dict]:
        task_identifier = "XasTemperaturePlot"
        output_directory = directories.get_processed_subdir(filename, "temperature")
        return [
            {
                "task_identifier": task_identifier,
                "name": "output_directory",
                "value": output_directory,
            },
            {
                "task_identifier": task_identifier,
                "name": "extend_plotrange_left",
                "value": self.extend_plotrange_left,
            },
            {
                "task_identifier": task_identifier,
                "name": "extend_plotrange_right",
                "value": self.extend_plotrange_right,
            },
            {
                "task_identifier": task_identifier,
                "name": "two_color_difference",
                "value": self.two_color_difference,
            },
            {
                "task_identifier": task_identifier,
                "name": "dpi",
                "value": self.dpi,
            },
        ]

    def _get_fit_inputs(self) -> List[dict]:
        if not self.refit:
            return []
        task_identifier = "PlanckRadianceFit"
        return [
            {
                "task_identifier": task_identifier,
                "name": "wavelength_min",
                "value": self.wavelength_min,
            },
            {
                "task_identifier": task_identifier,
                "name": "wavelength_max",
                "value": self.wavelength_max,
            },
        ]


class Plotter(WithFlintAccess):
    def __init__(self) -> None:
        super().__init__()
        self._tasks = list()

    def purge_tasks(self) -> int:
        """Remove references to tasks that have finished."""
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        """Kill all tasks."""
        gevent.killall(self._tasks)
        return self.purge_tasks()

    def handle_workflow_result(self, *args, **kwargs) -> None:
        """Handle workflow results in a background task"""
        self._spawn(self._handle_workflow_result, *args, **kwargs)

    def _handle_workflow_result(self, future, timeout: int = 60):
        try:
            results = future.get(timeout=timeout)
            if not results:
                return
            filenames = results["filenames"]
            directory = os.path.dirname(filenames[0])

            plot = self._get_plot()
            plot.select_directory(directory)
        except Exception as e:
            logger.exception(e)

    def _spawn(self, *args, **kw):
        task = gevent.spawn(*args, **kw)
        self._tasks.append(task)
        self.purge_tasks()

    def _get_plot(self) -> TemperaturePlot:
        return super()._get_plot("Temperature", TemperaturePlot)



import sys
import shutil
import tempfile
from pathlib import Path
from typing import Generator
from contextlib import contextmanager

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


@contextmanager
def resource_path(*args) -> Generator[Path, None, None]:
    """The resource is specified relative to `blissoda.resources`.

    .. code-block:: python

        with resource_path("exafs", "exafs.ows") a path:
            ...
    """
    source = importlib_resources.files(__name__).joinpath(*args)
    with importlib_resources.as_file(source) as path:
        yield path


def resource_filename(*args) -> str:
    """The resource is specified relative to `blissoda.resources`.

    .. code-block:: python

        filename = resource_filename("exafs", "exafs.ows")
    """
    with resource_path(*args) as path:
        if not path.exists():
            return str(path)  # resource does not exist
    if not path.exists():
        with resource_path(*args) as path:
            # resource was extract from zip: copy for persistency
            path_copy = Path(tempfile.mkdtemp()) / path.name
            shutil.copyfile(str(path), str(path_copy))
            path = path_copy
    return str(path)



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
from typing import Optional, List
from ..xrpd.processor import XrpdProcessor
from ..persistent.parameters import ParameterInfo


try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


class Id09XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("thickness", category="Corrections"),
    ],
):
    def __init__(self, **defaults) -> None:
        if setup_globals is None:
            raise ImportError("requires a bliss session")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 1024,
                "unit": "q_A^-1",
                "rot1": 0,
                "rot2": 0,
                "rot3": 0,
                "binning": 2,
            },
        )
        defaults.setdefault("thickness", 1)
        super().__init__(**defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(self, lima_name: str) -> dict:
        integration_options = self.integration_options

        if integration_options:
            integration_options = integration_options.to_dict()
        else:
            integration_options = dict()

        integration_options.setdefault("energy", setup_globals.get_xray_energy())
        integration_options.setdefault("binning", 2)

        has_rotation = (
            integration_options.get("rot1")
            or integration_options.get("rot2")
            or integration_options.get("rot3")
        )

        center = integration_options.pop("center", None)
        if center is not None:
            if has_rotation:
                print(
                    "WARNING: 'center' will be disregarded because rotations are not all zero."
                )
            else:
                poni1, poni2 = setup_globals.get_poni(center)
                integration_options["poni1"] = poni1
                integration_options["poni2"] = poni2

        return integration_options

    def master_output_filename(self, scan) -> str:
        filename = super().master_output_filename(scan)
        filename, ext = os.path.splitext(filename)
        return f"{filename}_oda{ext}"

    def external_output_filename(self, scan, lima_name: str) -> Optional[str]:
        filename = super().external_output_filename(scan, lima_name)
        if filename:
            filename, ext = os.path.splitext(filename)
            return f"{filename}_oda{ext}"

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        inputs.append(
            {
                "task_identifier": "SensorCorrection",
                "name": "sensor_thickness",
                "value": self.thickness,
            },
        )
        return inputs

    def get_submit_arguments(self, scan, lima_name) -> dict:
        kwargs = super().get_submit_arguments(scan, lima_name)
        kwargs["outputs"] = [{"task_identifier": "IntegrateBlissScanWithoutSaving"}]
        kwargs["merge_outputs"] = True
        return kwargs



from typing import Optional

try:
    from bliss import setup_globals
    from bliss.physics import units
except ImportError:
    units = None
    setup_globals = None

from ..exafs.plotter import ExafsPlotter
from ..resources import resource_filename


class DemoExafsPlotter(ExafsPlotter):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("workflow", resource_filename("exafs", "exafs.ows"))
        defaults.setdefault("_scan_type", "any")
        counters = defaults.setdefault("_counters", dict())
        energy_unit = setup_globals.energy.unit or "eV"
        counters.setdefault(
            "any",
            {
                "mu_name": "mu",
                "energy_name": "energy",
                "energy_unit": energy_unit,
            },
        )
        super().__init__(**defaults)

    def _scan_type_from_scan(self, scan) -> Optional[str]:
        return "any"

    def run(self, expo=0.003):
        e0 = 8800  # eV
        e1 = 9600  # eV
        step_size = 0.5  # eV
        intervals = int((e1 - e0) / step_size) - 1

        from_unit = "eV"
        to_unit = self.counters["energy_unit"]

        if units:
            e0 = (e0 * units.ur(from_unit)).to(to_unit).magnitude
            e1 = (e1 * units.ur(from_unit)).to(to_unit).magnitude
        else:
            assert (
                from_unit == to_unit
            ), f"counters energy unit is '{to_unit}' instead of 'eV'"

        scan = setup_globals.ascan(
            setup_globals.energy, e0, e1, intervals, expo, setup_globals.mu, run=False
        )
        super().run(scan)


if setup_globals is None:
    exafs_plotter = None
else:
    exafs_plotter = DemoExafsPlotter()
    exafs_plotter.refresh_period = 0.5



import json
import logging
from typing import Optional, Mapping

from ...exceptions import VersionError

# Note: import subscriber first because it might require patching
try:
    # blissdata >=1
    from .subscriberv1 import scan_iterator
except VersionError as exc:
    try:
        # blissdata >0.3.3,<1 (unreleased, branch id31_2.0)
        from .subscribervid31 import scan_iterator
    except VersionError:
        try:
            # blissdata <=0.3.3
            from .subscriberv0 import scan_iterator
        except VersionError:  # noqa F841
            _EXC = exc

            def scan_iterator(*args, **kw):
                raise _EXC


from ewoksjob.client import submit

logger = logging.getLogger(__name__)


def submit_scan_workflow(workflow=None, **kwargs) -> Optional[str]:
    if not workflow:
        return
    future = submit(args=(workflow,), kwargs=kwargs)
    return future.task_id


def main(args) -> None:
    for filename, scan_nb, workflows in scan_iterator(args.session):
        for wfname, nxprocess in workflows.items():
            if not isinstance(nxprocess, Mapping):
                continue
            try:
                job_id = submit_scan_workflow(
                    **json.loads(nxprocess["configuration"]["data"])
                )
            except Exception:
                logger.exception(
                    f"Error when submitting workflow '{wfname}' for scan {scan_nb} of file '{filename}'"
                )
            else:
                if job_id is not None:
                    logger.info(
                        f"Submitted workflow '{wfname}' (JOB ID {job_id}) for scan {scan_nb} of file '{filename}'"
                    )



import numpy
import pytest

try:
    from ..persistent.ordereddict import PersistentOrderedDict
except TypeError:
    # bliss not installed
    PersistentOrderedDict = None
from ..persistent.ndarray import PersistentNdArray
from ..persistent.parameters import WithPersistentParameters


def test_persistent_parameters(mock_persistent):
    class MyParameters(WithPersistentParameters, parameters=["a", "b"]):
        def __init__(self, **defaults) -> None:
            defaults.setdefault("a", 1)
            super().__init__(**defaults)

    parameters = MyParameters()

    expected = {"a": 1}
    assert mock_persistent == expected
    assert parameters.a == 1
    assert parameters.b is None

    parameters.a = 2
    expected["a"] = 2
    assert mock_persistent == expected
    assert parameters.a == 2

    parameters = MyParameters()
    assert parameters.a == 2

    parameters.a = None
    expected["a"] = None
    assert mock_persistent == {"a": None}
    assert parameters.a is None
    assert parameters.b is None


def test_persistent_parameters_dict(mock_persistent):
    class MyParameters(WithPersistentParameters, parameters=["python_dict"]):
        def __init__(self, **defaults) -> None:
            defaults.setdefault("python_dict", dict())
            super().__init__(**defaults)

    parameters = MyParameters()
    expected = {}

    assert mock_persistent == {"python_dict": expected}
    assert parameters.python_dict == expected

    parameters.python_dict["a"] = 2
    parameters.python_dict["fix"] = -1
    expected["python_dict"] = {"a": 2, "fix": -1}
    assert mock_persistent == expected
    assert parameters.python_dict == expected["python_dict"]

    parameters.python_dict["a"] = {"x": 1, "fix": -2}
    expected["python_dict"]["a"] = {"x": 1, "fix": -2}
    assert mock_persistent == expected
    assert parameters.python_dict["a"]["x"] == 1

    parameters.python_dict["a"]["x"] = {"y": 2, "fix": -3}
    expected["python_dict"]["a"]["x"] = {"y": 2, "fix": -3}
    assert mock_persistent == expected
    assert parameters.python_dict["a"]["x"]["y"] == 2

    parameters.python_dict["a"]["x"]["y"] = 3
    expected["python_dict"]["a"]["x"]["y"] = 3
    assert mock_persistent == expected
    assert parameters.python_dict["a"]["x"]["y"] == 3


def test_persistent_ndarray(mock_bliss):
    python_arr = list()
    persistent_arr = PersistentNdArray("test")
    with pytest.raises(IndexError):
        persistent_arr[0]
    with pytest.raises(IndexError):
        persistent_arr[-1]
    numpy.testing.assert_array_equal(python_arr, persistent_arr[()])

    add = numpy.random.uniform(low=0, high=1, size=10)
    python_arr.append(add)
    persistent_arr.append(add)
    python_arr_copy = numpy.array(python_arr)
    numpy.testing.assert_array_equal(python_arr_copy[0], persistent_arr[0])
    numpy.testing.assert_array_equal(python_arr_copy[-1], persistent_arr[-1])
    numpy.testing.assert_array_equal(python_arr_copy, persistent_arr[()])

    add = numpy.random.uniform(low=0, high=1, size=(2, 10))
    python_arr.extend(add)
    persistent_arr.extend(add)
    python_arr_copy = numpy.array(python_arr)
    numpy.testing.assert_array_equal(python_arr_copy[0], persistent_arr[0])
    numpy.testing.assert_array_equal(python_arr_copy[-1], persistent_arr[-1])
    numpy.testing.assert_array_equal(python_arr_copy, persistent_arr[()])

    numpy.testing.assert_array_equal(python_arr_copy[2, 5:6], persistent_arr[2, 5:6])


def test_persistent_ordered_dict(mock_bliss):
    python_dict = dict()
    persistent_dict = PersistentOrderedDict("test")
    python_dict["string"] = "abc"
    persistent_dict["string"] = "abc"
    python_dict["number"] = 123
    persistent_dict["number"] = 123
    python_dict["list"] = [123, 456]
    persistent_dict["list"] = [123, 456]
    python_dict["dict"] = {"key": "value"}
    persistent_dict["dict"] = {"key": "value"}
    assert python_dict == persistent_dict.get_all()



try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from blissoda.demo.xrpd import xrpd_processor


def xrpd_demo(expo=0.2, npoints=10):
    xrpd_processor.enable(setup_globals.difflab6)
    try:
        pct(
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
        setup_globals.loopscan(
            npoints,
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
    finally:
        xrpd_processor.disable()


def pct(*args, **kw):
    s = setup_globals.ct(*args, **kw)
    return xrpd_processor.on_new_scan(s)



from blissoda.demo.id22 import stscan_processor
from blissoda.demo.id22 import id22_xrpd_processor

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


def id22_stscan_demo(expo=0.2, npoints=10):
    stscan_processor.submit_workflows()


def id22_xrpd_demo(expo=0.2, npoints=10):
    id22_xrpd_processor.enable(setup_globals.difflab6)
    try:
        setup_globals.loopscan(
            npoints,
            expo,
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        )
    finally:
        id22_xrpd_processor.disable()



"""Flint-side for Flint ID24 plots"""

from ..flint import capture_errors
from ..gui.image_gallery import ImageGalleryViewer


class TemperatureWidget(ImageGalleryViewer):
    @capture_errors
    def select_directory(self, directory: str) -> None:
        return super().select_directory(directory)



import logging
from typing import Iterator, Tuple, Dict

from ...exceptions import VersionError

try:
    from blissdata.beacon.data import BeaconData
    from blissdata.redis_engine.store import DataStore
    from blissdata.redis_engine.scan import Scan
except ImportError as e:
    raise VersionError(str(e)) from e


logger = logging.getLogger(__name__)


def scan_iterator(session_name) -> Iterator[Tuple[str, int, Dict]]:
    logger.info(f"Started listening to Bliss session '{session_name}'")

    redis_url = BeaconData().get_redis_data_db()
    data_store = DataStore(redis_url)
    since = data_store.get_last_scan_timetag()
    while True:
        since, key = data_store.get_next_scan(since=since)
        scan = data_store.load_scan(key, scan_cls=Scan)
        if scan.session != session_name:
            continue
        if scan.info.get("is-scan-sequence") or scan.info.get("group"):
            continue
        workflows = scan.info.get("workflows")
        if not workflows:
            continue
        filename = scan.info.get("filename")
        yield filename, scan.number, workflows



from typing import Tuple, List, Optional, Callable

import numpy
from silx.io import h5py_utils
from silx.utils.retry import RetryError
from silx.utils.retry import RetryTimeoutError

from ..persistent.ordereddict import PersistentOrderedDict
from ..persistent.ndarray import PersistentNdArray


def get_redis_store(redis_key: str) -> PersistentOrderedDict:
    return PersistentOrderedDict(redis_key)


def add_plot(
    redis_store: PersistentOrderedDict,
    scan_name: str,
    lima_name: str,
    signal_ndim: int,
    xlabel: str,
    xvalues: numpy.ndarray,
    hdf5_url: Optional[str] = None,
) -> Tuple[str, dict]:
    """Add a new curve plot."""
    plot_key = f"{scan_name}:{lima_name}"
    data_key = f"{redis_store.name}:{plot_key}"
    if hdf5_url is None:
        hdf5_url = ""
    plot_info = {
        "scan_name": scan_name,
        "lima_name": lima_name,
        "xlabel": xlabel,
        "signal_ndim": signal_ndim,
        "data_key": data_key,
        "hdf5_url": hdf5_url,
    }
    redis_store[plot_key] = plot_info
    PersistentNdArray(data_key).append(xvalues)
    return plot_key, plot_info


def remove_plots(
    redis_store: PersistentOrderedDict,
    max_len: int,
    remove_from_flint: Callable,
) -> List[Tuple[str, dict]]:
    """Remove plots when there are more than `max_len`."""
    all_plots = list(redis_store.items())
    if max_len > 0:
        remove = list()
        keep_scans = set()
        for plot_key, plot_info in all_plots[::-1]:
            scan_name = plot_info["scan_name"]
            if len(keep_scans) == max_len and scan_name not in keep_scans:
                remove.append((plot_key, plot_info))
            else:
                keep_scans.add(scan_name)
    else:
        remove = all_plots
    if remove:
        for plot_key, plot_info in remove:
            remove_from_flint(plot_key, plot_info)
            PersistentNdArray(plot_info["data_key"]).remove()
            redis_store.remove(plot_key)
    return remove


def add_data(plot_key: str, plot_info: dict, points: numpy.ndarray) -> str:
    """Append data to an existing plot."""
    PersistentNdArray(plot_info["data_key"]).extend(points)
    return plot_key


def get_curve_data(
    redis_store: PersistentOrderedDict, plot_key: str, **retry_options
) -> Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray], dict]:
    """Get the data from the results of the last processed point of the scan."""
    plot_info = redis_store[plot_key]
    data = PersistentNdArray(plot_info["data_key"])
    try:
        x = data[0]
    except IndexError:
        return None, None, plot_info

    if plot_info["hdf5_url"]:
        try:
            if plot_info["signal_ndim"] == 1:
                idx = tuple()
            else:
                idx = -1
        except KeyError:
            # Plot produced by an old version of blissoda
            if plot_info["ct"]:
                idx = tuple()
            else:
                idx = -1
        try:
            y = _get_data_from_file(plot_info["hdf5_url"], idx=idx, **retry_options)
        except RetryTimeoutError:
            y = None
        return x, y, plot_info
    else:
        try:
            y = data[-1]
        except IndexError:
            return None, None, plot_info

        if y.ndim == 2:
            return x, y[-1], plot_info
        return x, y, plot_info


def append_image_data(
    redis_store: PersistentOrderedDict,
    plot_key: str,
    current_data: Optional[numpy.ndarray] = None,
    **retry_options,
) -> Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray], dict]:
    """Add new data (if any) to the current data (if any)"""
    plot_info = redis_store[plot_key]
    data = PersistentNdArray(plot_info["data_key"])
    try:
        x = data[0]
    except IndexError:
        return None, None, plot_info

    if plot_info["hdf5_url"]:
        if current_data is None:
            idx = tuple()
        else:
            idx = slice(len(current_data), None)
        try:
            y = _get_data_from_file(plot_info["hdf5_url"], idx=idx, **retry_options)
        except RetryTimeoutError:
            y = None
        else:
            if current_data is not None:
                y = numpy.vstack([current_data, y])
    else:
        y = data[1:]

    return x, y, plot_info


@h5py_utils.retry()
def _get_data_from_file(hdf5_url: str, idx=tuple()):
    filename, dsetname = hdf5_url.split("::")
    with h5py_utils.File(filename) as root:
        try:
            return root[dsetname][idx]
        except KeyError as e:
            raise RetryError(str(e))



r"""Optimize exposure conditions (time and attenuator) based
on a measurement

* $I_m$: measured diffracted intensity (maximum pixel value)
* $I_d$: desired diffracted intensity (maximum pixel value)
* $R$: diffraction count rate (Hz, maximum pixel value per second)
* $T(E,n)$: transmission at energy $E$ and attenuator position $n$

.. math::

    \begin{align}
    I_m &= R * tframe * nframe_m * T(E,n_m) \\
    I_d &= R * tframe * nframe_d * T(E,n_d)
    \end{align}

Solve the following equation to $n_d$ and $nframe_d$

.. math::

    \frac{nframe_d * T(E,n_d)} = \frac{I_d * nframe_m * T(E,n_m)}{I_m}
"""

from typing import Union, Optional, NamedTuple, List
from contextlib import contextmanager
from functools import lru_cache

import numpy
from scipy.ndimage import gaussian_filter

try:
    from bliss import setup_globals

    try:
        from blissdata.lima import image_utils

        lima_image = None
    except ImportError:
        image_utils = None
        from blissdata.data import lima_image
except ImportError:
    setup_globals = None
    lima_image = None
    image_utils = None

try:
    from id31 import attenuator as id31_attenuator
except ImportError:
    id31_attenuator = None


class ExposureCondition(NamedTuple):
    att_position: int
    expo_time: float


def optimize_exposure_condition(
    detector,
    tframe: float = 0.2,
    default_att_position: Optional[int] = None,
    desired_counts: float = 1e5,
    dynamic_range: int = 1 << 20,
    min_counts_per_frame: float = 1000,
    nframes_measure: int = 1,
    nframes_default: int = 3,
    reduce_desired_deviation: bool = True,
    expose_with_integral_frames: bool = False,
    maxint_sigma: float = 0,
) -> ExposureCondition:
    """Optimize the attenuator and return the optimal exposure time.

    :param detector: lima detector object
    :param tframe: time of a single frame in accumulation mode
    :param default_att_position: start the optimization with this attenuator position
    :param desired_counts: we want to be close to this maximum pixel value
    :param dynamic_range: full well of the detector
    :param min_counts_per_frame: we need at least these counts to even start optimizing
    :param nframes_measure: number of frames used for optimization measurements
    :param nframes_default: try modify attenuators for this number of frames
    :param reduce_desired_deviation: reduce deviation from the desired counts
    :param expose_with_integral_frames: exposure time is n x frame time with n integral or not
    :param maxint_sigma: gaussian smoothing sigma in pixels before getting the maximum frame intensity
    :returns: optimal exposure condition
    """
    Imax_perframe = dynamic_range - 1
    Id = desired_counts
    att_position_max = 31
    att_position_start = _get_attenuator_position()

    with _diode_range_context():
        if default_att_position is None:
            att_position = att_position_start
        else:
            att_position = default_att_position
            if att_position != att_position_start:
                _set_attenuator_position(att_position)

        # Make sure we are BELOW the dynamic range of the detector
        Im_perframe = _get_max_intensity_per_frame(detector, tframe, nframes_measure)
        while Im_perframe >= Imax_perframe and att_position < att_position_max:
            att_position += 1
            _set_attenuator_position(att_position)
            Im_perframe = _get_max_intensity_per_frame(
                detector, tframe, nframes_measure
            )

        if Im_perframe >= Imax_perframe:
            # We are at full attenuation and full dynamic range.
            # Decreasing the frame time would be an option but
            # ID31 keeps it fixed.
            print(
                f"Optimized exposure: expo_time={tframe} sec, attenuator={att_position}"
            )
            return tframe

        # Make sure we have some counts to make the calculations
        # further on more reliable
        while Im_perframe < min_counts_per_frame and att_position:
            att_position -= 1
            _set_attenuator_position(att_position)
            _prev = Im_perframe
            Im_perframe = _get_max_intensity_per_frame(
                detector, tframe, nframes_measure
            )
            if Im_perframe >= Imax_perframe:
                att_position += 1
                Im_perframe = _prev
                break

        # Calculate transmissions T(E,n_d) of all attenuators
        transmissions = _attenuator_transmisions(_get_energy_value())

        # Calculate the conditions to reach the desired counts
        condition = _calculate_optional_conditions(
            transmissions,
            att_position,
            Im_perframe,
            Id,
            tframe=tframe,
            nframes_default=nframes_default,
            reduce_desired_deviation=reduce_desired_deviation,
            expose_with_integral_frames=expose_with_integral_frames,
        )

        _set_attenuator_position(condition.att_position)
        return condition


def optimal_exposure_conditions(
    mot,
    start,
    stop,
    intervals,
    detector,
    tframe: float = 0.2,
    nframes_measure: int = 1,
    nframes_default: int = 3,
    desired_counts: float = 1e5,
    reduce_desired_deviation: bool = True,
    expose_with_integral_frames: bool = False,
    maxint_sigma: float = 2,
) -> List[ExposureCondition]:
    """Return the optimal attenuator and exposure time for each motor position.
    Measurements are done at the current attenuator position.

    :param mot:
    :param start:
    :param stop:
    :param intervals:
    :param detector: lima detector object
    :param tframe: time of a single frame in accumulation mode
    :param desired_counts: we want to be close to this maximum pixel value
    :param nframes_measure: number of frames used for optimization measurements
    :param nframes_default: try modify attenuators for this number of frames
    :param reduce_desired_deviation: reduce deviation from the desired counts
    :param expose_with_integral_frames: exposure time is n x frame time with n integral or not
    :param maxint_sigma: gaussian smoothing sigma in pixels before getting the maximum frame intensity
    :returns: list of optimal exposure conditions
    """
    Id = desired_counts
    att_position = _get_attenuator_position()

    Im_perframe_per_position = _get_max_intensity_per_frame_per_position(
        mot,
        start,
        stop,
        intervals,
        detector,
        tframe=tframe,
        nframes=nframes_measure,
        sigma=maxint_sigma,
    )

    transmissions = _attenuator_transmisions(_get_energy_value())

    return [
        _calculate_optional_conditions(
            transmissions,
            att_position,
            Im_perframe,
            Id,
            tframe=tframe,
            nframes_default=nframes_default,
            reduce_desired_deviation=reduce_desired_deviation,
            expose_with_integral_frames=expose_with_integral_frames,
        )
        for Im_perframe in Im_perframe_per_position
    ]


def _calculate_optional_conditions(
    transmissions,
    att_position,
    Im_perframe,
    Id,
    tframe: float = 0.2,
    nframes_default: int = 3,
    reduce_desired_deviation: bool = True,
    expose_with_integral_frames: bool = False,
) -> ExposureCondition:
    # Calculate all possible intensities with nframe_d=nframes_default
    #  I = R * tframe * nframes_default * T(E,n_d)
    Rtframe = Im_perframe / transmissions[att_position]  # R * tframe
    Ichoices = Rtframe * nframes_default * transmissions
    att_position = numpy.argmin(abs(Ichoices - Id))

    if att_position != 0:
        if reduce_desired_deviation:
            # Reduce the deviation from Id by solving this to nframes_d
            #  I_d = R * tframe * nframes_d * T(E,n_d)
            nframe_d = Id / (Rtframe * transmissions[att_position])
            if expose_with_integral_frames:
                expo_time = _round_number(nframe_d) * tframe
            else:
                expo_time = _round_number(nframe_d * tframe, ndecimals=1)
        else:
            expo_time = nframes_default * tframe
    else:
        # No attenuator: increase exposure time
        #  I_d = R * tframe * nframes
        nframe_d = Id / Rtframe
        if expose_with_integral_frames:
            expo_time = _round_number(nframe_d) * tframe
        else:
            expo_time = _round_number(nframe_d * tframe, ndecimals=1)

    expected_max_counts = Rtframe / tframe * expo_time * transmissions[att_position]
    print(
        f"Optimal exposure conditions: expo_time={expo_time} sec, attenuator={att_position}, expected max counts={expected_max_counts}"
    )
    return ExposureCondition(att_position=att_position, expo_time=expo_time)


def _set_attenuator_position(att_position: int) -> None:
    """SiO2 thickness (cm) ~= 1.25 * att_position. Does not adapt the diode ranges accordingly."""
    setup_globals.atten.bits = att_position


def _get_attenuator_position() -> int:
    return setup_globals.atten.bits


def _get_energy_value() -> float:
    """keV"""
    return setup_globals.energy.position


@lru_cache(maxsize=1)
def _attenuator_transmisions(energy: float):
    att_position_max = 31
    return id31_attenuator.SiO2trans(energy, numpy.arange(att_position_max + 1))


@contextmanager
def _diode_range_context():
    """Adapt the diode ranges according to the attenuator position."""
    try:
        yield
    finally:
        setup_globals.att(setup_globals.atten.bits)


def _round_number(nframes: Union[float, int], ndecimals: int = 0) -> Union[float, int]:
    if ndecimals:
        m = 10**ndecimals
        return max(int(nframes * m + 0.5) / m, 1 / m)
    else:
        return max(int(nframes + 0.5), 1)


def _get_max_intensity_per_frame(
    detector,
    tframe: float = 0.2,
    nframes: int = 1,
    sigma: float = 0,  # rockit is most likely enabled
) -> float:
    setup_globals.ct(tframe * nframes, detector)

    # This is slower:
    # setup_globals.fshopen()
    # setup_globals.limatake(tframe, nbframes=nframes)
    # setup_globals.fshclose()

    if image_utils is not None:
        frame = image_utils.read_video_last_image(detector.proxy).array
    else:
        frame, _ = lima_image.read_video_last_image(detector.proxy)
    maxint = _get_frame_max_intensity(frame, sigma=sigma) / nframes
    return maxint


def _get_max_intensity_per_frame_per_position(
    mot,
    start,
    stop,
    intervals,
    detector,
    tframe: float = 0.2,
    nframes: int = 1,
    sigma: float = 2,  # rockit is cannot be enabled
) -> List[float]:
    count_time = tframe * nframes
    scan = setup_globals.ascan(
        mot, start, stop, intervals, count_time, detector, save=False
    )

    data = scan.get_data()
    frames = data[detector.name + ":image"]
    maxint = [
        _get_frame_max_intensity(frame, sigma=sigma) / nframes for frame in frames
    ]
    return maxint


def _get_frame_max_intensity(frame, sigma: float = 2) -> float:
    if sigma > 0:
        return gaussian_filter(frame, sigma=sigma).max()
    return frame.max()



from contextlib import ExitStack
from contextlib import contextmanager
from unittest.mock import patch

from . import setup_globals


from . import attenuator

try:
    import blissdata.lima.image_utils  # noqa F401
    from . import image_utils

    lima_image = None
except ImportError:
    from . import lima_image

    image_utils = None


@contextmanager
def mock_id31():
    with ExitStack() as stack:
        ctx = patch("blissoda.id31.optimize_exposure.setup_globals", new=setup_globals)
        stack.enter_context(ctx)
        ctx = patch("blissoda.id31.optimize_exposure.id31_attenuator", new=attenuator)
        stack.enter_context(ctx)
        if image_utils is None:
            ctx = patch("blissoda.id31.optimize_exposure.lima_image", new=lima_image)
            stack.enter_context(ctx)
        else:
            ctx = patch("blissoda.id31.optimize_exposure.image_utils", new=image_utils)
            stack.enter_context(ctx)
        yield



"""User API for HDF5 conversion on the Bliss repl"""

import os
import re
from typing import List, Dict

from ewoksjob.client import submit

try:
    from bliss import current_session
except ImportError:
    current_session = None

from ..utils import trigger
from ..persistent.parameters import ParameterInfo
from ..persistent.parameters import WithPersistentParameters
from ..resources import resource_filename


class Id12Hdf5ToAsciiConverter(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("_enabled"),
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("external_proposal_outdir", category="parameters"),
        ParameterInfo("inhouse_proposal_outdir", category="parameters"),
        ParameterInfo("test_proposal_outdir", category="parameters"),
        ParameterInfo("retry_timeout", category="data access"),
    ],
):
    def __init__(self, **defaults) -> None:
        if current_session is None:
            raise ImportError("bliss")
        defaults.setdefault("_enabled", False)
        root_dir = "/data/id12/inhouse"
        defaults.setdefault(
            "external_proposal_outdir", os.path.join(root_dir, "EXTERNAL")
        )
        defaults.setdefault(
            "inhouse_proposal_outdir", os.path.join(root_dir, "INHOUSE2")
        )
        defaults.setdefault("test_proposal_outdir", os.path.join(root_dir, "NOBACKUP"))
        defaults.setdefault("workflow", resource_filename("id12", "convert.json"))
        super().__init__(**defaults)
        self._sync_scan_metadata()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        categories["status"] = {"Enabled": self._enabled}
        return categories

    def enable(self, *detectors) -> None:
        self._enabled = True
        self._sync_scan_metadata()

    def disable(self) -> None:
        self._enabled = False
        self._sync_scan_metadata()

    def _sync_scan_metadata(self) -> None:
        if self._enabled:
            workflows_category = trigger.register_workflow_category(timing="END")
            workflows_category.set("processing", self.on_new_scan_metadata)
        else:
            trigger.unregister_workflow_category()

    def on_new_scan_metadata(self, scan) -> None:
        if not self.scan_requires_processing(scan):
            return
        kwargs = self.get_submit_arguments(scan)
        _ = submit(args=(self.workflow,), kwargs=kwargs)

    def scan_requires_processing(self, scan) -> bool:
        return scan.scan_info["filename"] and scan.scan_info["save"]

    def get_submit_arguments(self, scan) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_inputs(self, scan) -> List[dict]:
        task_identifier = "Hdf5ToAscii"

        filename = scan.scan_info["filename"]
        output_filename = self.output_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")

        return [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": output_filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "scan_numbers",
                "value": [scan_nb],
            },
        ]

    def output_filename(self, scan) -> str:
        proposal = current_session.scan_saving.proposal.name

        # Proposal directory is upper case and "-" between letters and digits
        matches = re.findall(r"[A-Za-z]+|\d+", proposal.upper())
        proposal_dir = "-".join(matches)

        # Handle special cases
        proposal_dir = proposal_dir.replace("IH", "IH-")
        proposal_dir = re.sub(r"ID-(\d{2})(.+)", r"ID\1-\2", proposal_dir)

        # Select directory for the ASCII files
        if current_session.scan_saving.proposal_type == "inhouse":
            dirname = self.inhouse_proposal_outdir
        elif current_session.scan_saving.proposal_type == "tmp":
            dirname = self.test_proposal_outdir
        elif proposal_dir.startswith("IH") or proposal_dir.startswith("BLC"):
            dirname = self.inhouse_proposal_outdir
        else:
            dirname = self.external_proposal_outdir

        filename = scan.scan_info["filename"]
        collection_dataset = os.path.splitext(os.path.basename(filename))[0]
        return os.path.join(dirname, proposal_dir, collection_dataset + ".dat")



from typing import Optional

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from ..exafs.plotter import ExafsPlotter


class Id24ExafsPlotter(ExafsPlotter):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("workflow", "/users/opid24/ewoks/online.ows")
        defaults.setdefault("_scan_type", "escan")
        counters = defaults.setdefault("_counters", dict())
        counters.setdefault(
            "escan",
            {
                "mu_name": "mu_trans",
                "energy_name": "energy_enc",
                "energy_unit": "keV",
            },
        )
        super().__init__(**defaults)

    def _scan_type_from_scan(self, scan) -> Optional[str]:
        return "escan"

    def run(self, scan, **kw):
        super().run(scan, filename=setup_globals.SCAN_SAVING.filename, **kw)



from pprint import pprint

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from ..id14.converter import Id14Hdf5ToSpecConverter


class DemoId14Hdf5ToSpecConverter(Id14Hdf5ToSpecConverter):
    def on_new_scan_metadata(self, scan) -> None:
        kwargs = self.get_submit_arguments(scan)
        print("Workflow", self.workflow)
        print(" -> arguments")
        pprint(kwargs)


if setup_globals is None:
    id14_converter = None
else:
    id14_converter = DemoId14Hdf5ToSpecConverter()



from typing import Optional, List

from ..xrpd.processor import XrpdProcessor
from ..persistent.parameters import ParameterInfo
from ..resources import resource_filename
from ..utils.pyfai import read_config


class Id13XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    DEFAULT_WORKFLOW: Optional[str] = resource_filename(
        "id13", "integrate_scan_with_saving.json"
    )
    DEFAULT_LIMA_URL_TEMPLATE: Optional[str] = (
        "{dirname}/scan{scan_number_as_str}/{images_prefix}{{file_index}}.h5::/entry_0000/measurement/data"
    )

    def __init__(self, **defaults) -> None:
        defaults.setdefault("save_scans_separately", True)
        super().__init__(**defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        if self.pyfai_config:
            integration_options = read_config(filename=self.pyfai_config)
        else:
            integration_options = dict()
        if self.integration_options:
            integration_options.update(self.integration_options.to_dict())
        return integration_options

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        eval_dict = {"img_acq_device": lima_name, "scan_number": scan.scan_number}
        images_prefix = scan.scan_saving.eval_template(
            scan.scan_saving.images_prefix, eval_dict=eval_dict
        )
        original_template_args = self.lima_url_template_args.to_dict()
        self.lima_url_template_args["images_prefix"] = images_prefix
        self.lima_url_template_args["scan_number_as_str"] = scan.scan_number
        try:
            return super().get_integrate_inputs(scan, lima_name, task_identifier)
        finally:
            self.lima_url_template_args = original_template_args
